# -*- coding: utf-8 -*-
"""
    __author__ = 'Zehra Sarica'
    __email__ = ['sarica16@itu.edu.tr','zehraacar559@gmail.com']
"""

FONT_FAMILY = "Times New Roman"

FONT_STYLES = {
    "title": {"fontsize": 16, "fontfamily": FONT_FAMILY},
    "xlabel": {"fontsize": 14, "fontfamily": FONT_FAMILY},
    "ylabel": {"fontsize": 14, "fontfamily": FONT_FAMILY},
    "zlabel": {"fontsize": 14, "fontfamily": FONT_FAMILY},
    "xtick": {"labelsize": 14, "fontfamily": FONT_FAMILY},
    "xtick_medium": {"labelsize": 12, "fontfamily": FONT_FAMILY},
    "xtick_small": {"labelsize": 10, "fontfamily": FONT_FAMILY},
    "xtick_small_small": {"labelsize": 8, "fontfamily": FONT_FAMILY},
    "ytick": {"labelsize": 14, "fontfamily": FONT_FAMILY},
    "tick": {"labelsize": 14, "fontfamily": FONT_FAMILY},
    "legend": {
        "fontsize": 14,
        "title_fontsize": 14,
        "fontfamily": FONT_FAMILY,
    },
    "colorbar": {
        "label": {"fontsize": 14, "fontfamily": FONT_FAMILY},
        "tick_params": {"labelsize": 14, "fontfamily": FONT_FAMILY},
        "tick_params_label": {"labelsize": 14}
    },
}

EDGE_COLOR = '#1f4e79'
COLOR_PALETTE = ['red', 'gray', 'blue', 'green', 'purple', 'brown', 'teal', 'gold', 'black', 'magenta']
